//
// Example of MRT in OpenGL 
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#include	<glut.h>
#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"Data.h"
#include	"GlslProgram.h"
#include	"Framebuffer.h"
#include	"boxes.h"

Vector3D	eye   ( 7, 5, 7 );			// camera position
Vector3D	light ( 5, 0, 4 );			// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
int			activeMap = 0;
unsigned	diffMap;
unsigned	decalMap;

#define	IMAGE_SIZE	512

FrameBuffer	buffer  ( IMAGE_SIZE, IMAGE_SIZE, FrameBuffer :: depth32 );
GlslProgram	blinnProgram;

void reshape ( int w, int h );

/////////////////////////////////////////////////////////////////////////////////

void init ()
{
	glClearColor ( 0, 0, 0, 1 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void	drawQuad ( unsigned tex )
{
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glColor4f     ( 1, 1, 1, 1 );
    glBindTexture ( GL_TEXTURE_2D, tex );
	
	glEnable      ( GL_TEXTURE_2D );
    glBegin       ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1, 0 );
        glVertex2f   ( 1, 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( 1, 1 );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, 1 );

    glEnd   ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, 1, 0, 1, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix

	glDisable   ( GL_DEPTH_TEST );
	glDepthMask ( GL_FALSE );
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix

	glEnable    ( GL_DEPTH_TEST );
	glDepthMask ( GL_TRUE );
}

void display ()
{
	GLenum	buffers [] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
	
												// render for FP FBO
	buffer.bind ();

	glDrawBuffers ( 2, buffers );
	
	reshape ( buffer.getWidth (), buffer.getHeight () );
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

												// draw the light
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	glDisable          ( GL_TEXTURE_2D );
	glTranslatef       ( light.x, light.y, light.z );
	glColor4f          ( 1, 1, 1, 1 );
	glutSolidSphere    ( 0.1f, 15, 15 );
	glPopMatrix        ();

	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

	glBindTexture ( GL_TEXTURE_2D, diffMap );
	
	blinnProgram.bind ();
	
	drawBox ( Vector3D ( -1, -1, -2 ), Vector3D ( 1.7, 1.7, 1.7 ), decalMap );
    
	glTranslatef ( 2.5, 2.5, 1 );
	glRotatef    ( 15*angle, 0, 1, 0 );
	
	glutSolidTorus ( 0.7, 1.6, 20, 20 );
	
	glRotatef    ( 10*angle, 0, 0, 1 );
	
	drawBox ( Vector3D ( -2, -2, 1 ), Vector3D ( 1.5, 1.5, 1.5 ), diffMap );
	
	glEnable ( GL_TEXTURE_2D );
    
	blinnProgram.unbind ();
	buffer.unbind       ();

    glPopMatrix     ();
	
												// now process image (blur)
	startOrtho ();
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	drawQuad ( buffer.getColorBuffer ( activeMap ) );
	
	endOrtho ();
	
	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   											// factor all camera ops into projection matrix
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0.0, 0.0, 1.0 );		// up

   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
		
	if ( key == '0' )
		activeMap = 0;
	else
	if ( key == '1' )
		activeMap = 1;
}

void	animate ()
{
	angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

	light.x = 2*cos ( angle );
	light.y = 3*sin ( 1.4 * angle );
	light.z = 3 + 0.5 * sin ( angle / 3 );

	blinnProgram.bind ();
	blinnProgram.setUniformVector ( "eyePos",   Vector4D ( eye,   1 ) );
	blinnProgram.setUniformVector ( "lightPos", Vector4D ( light, 1 ) );
	blinnProgram.unbind ();

	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( IMAGE_SIZE, IMAGE_SIZE );


								// create window
	glutCreateWindow ( "Example of MRT in OpenGL" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init           ();
	initExtensions ();
	
	assertExtensionsSupported ( "GL_ARB_shading_language_100 GL_ARB_shader_objects ARB_color_buffer_float" );

	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}
	
    diffMap = createTexture2D ( true,  "../../Textures/oak.bmp" );
    decalMap = createTexture2D ( true,  "../../Textures/16.jpg" );

	if ( !blinnProgram.loadShaders ( "blinn.vsh", "blinn.fsh" ) )
	{
		printf ( "Error loading blinn-2 shaders:\n%s\n", blinnProgram.getLog ().c_str () );

		return 3;
	}

	blinnProgram.bind       ();
	blinnProgram.setTexture ( "diffMap", 0 );
	blinnProgram.unbind     ();
	
	GLenum	 format        = GL_RGBA8;
	unsigned screenMapMain = buffer.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );
	unsigned screenMapAdd  = buffer.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );
	
	buffer.create ();
	buffer.bind   ();
	
	if ( !buffer.attachColorTexture ( GL_TEXTURE_2D, screenMapMain, 0 ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.attachColorTexture ( GL_TEXTURE_2D, screenMapAdd, 1 ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();
	
    glutMainLoop ();

	return 0;
}
